import openpyxl
import pandas as pd
import numpy as np
import os
import json
import re
import warnings
import time
from statistics import mean
from pathlib import Path
warnings.simplefilter("ignore")

class Calculate:
    def __init__(self):
        self.statics = self.read_json()
        self.static_tables = self.statics['tables']
        self.days = self.statics['days']

        #self.step2fileread(self.statics['dir_step2'])
        #self.batchfileread(self.statics['dir_batch'])
        self.outfolderset(self.statics['out_folder'])
    
    def step2fileread(self, dir):
        self.dir_step2 = dir
        self.files1 = os.listdir(dir)
        
        print('\n', '#' * 10, 'FILES FOR STEP 2 FOUND', '#' * 10,'\n')
        for index, files in enumerate(self.files1):
            print(f'{index} -) {files}')
        print()
        
    def batchfileread(self, dir):
        self.dir_batch = dir
        self.files2 = os.listdir(dir)
        
        print('\n', '#' * 10, 'FILES FOR BATCH FOUND', '#' * 10,'\n')
        for index, files in enumerate(self.files2):
            print(f'{index} -) {files}')
        print()
        
    def outfolderset(self, dir):
        self.out_folder = dir
    
    def read_json(self):
        with open('settings/statics.json', 'r') as file:
            data = file.read()

        return json.loads(data)
        
    def get_samples(self, df):
    
        #print(df.columns)
        locs = df.apply(lambda xx: xx.count(), axis=1)
        n_tot = len(locs)
        nx = locs[locs == 0]
        n_10 = nx.index[0] if len(nx) > 0 else n_tot
        #print(n_10)
        
        df_input = df.loc[:,[i for i in df.columns if i.startswith("Average")]]
        df_input = df_input.iloc[:n_10,1:].div(df_input.iloc[:n_10,0], axis=0)
         
        #df_input = df.iloc[n_l0:n_tot,:].dropna(how='all', axis=1)
        #print(df_input)
        return df_input
    
    def colname_s2(self, i):
        return 'Count' if i < 1 else 'Count.{}'.format(i)
    
    def step2(self, stepf):
        well_length = 5
        
        ## SHEET
        wb = openpyxl.load_workbook(stepf)
        if len(wb.sheetnames) > 1:
            print("Sheet Names", wb.sheetnames, '\n')
            collist = sorted(list(set(wb.sheetnames)-set(['Control','Summary'])))
        
            tables = {x: pd.read_excel(stepf, sheet_name=x, skiprows=1, engine="openpyxl") for x in collist}
            totals = {'Well-{}'.format(j+1): [tables[x][self.colname_s2(i)].notna().sum() for i in range(well_length)] 
                        for j, x in enumerate(tables.keys())}
        
        else:
            table = pd.read_excel(stepf, skiprows=1, engine="openpyxl")
            limits = [(i*30, (i+1)*30 - 1) for i in range(6)]
            
            tables = {chr(65+i): table.iloc[:,l:h] for i, (l, h) in enumerate(limits)}
            totals = {'Well-{}'.format(j+1): [tables[x][self.colname_s2(i)].notna().sum() for i in range(j*well_length, (j+1)*well_length)] 
                        for j, x in enumerate(tables.keys())}
            
            #A0 = len(cstep01.iloc[:, :29]['Count'].dropna())
            
        ## WELLS
        wells = pd.DataFrame.from_dict(totals, orient='index', columns = ['Frame {}'.format(i+1) for i in range(well_length)])
        means = wells.mean(axis=1)
        wells['Average'] = means
        print("STEP 2 SUMMARY\n",  '-'*15+'\n', wells.to_string(), '\n', sep='')
        #print('\n', '#' * 10 + ' Step-2 Mean ' + '#' * 10, '\n')
        #print(*['Well-{}: {}\n'.format(i+1, mu) for i, mu in enumerate(means)], sep='')
        
        return (means, wells, tables)
    
    def batch(self, batch_filename, sheetname, control_average):
        Element = pd.read_excel(batch_filename, sheet_name = sheetname, engine="openpyxl", header=0)
        Element_input = self.get_samples(Element)
        
        Element_result = []
        
        nEl = len(Element_input)
        nElc = len(Element_input.columns)
        
        for i_1 in range(nEl):
            tmp_column = []
            for j_1 in range(nElc):
                tmp_column.append(
                    (control_average[j_1] - Element_input.iloc[i_1].iloc[j_1]) / (control_average[j_1])
                )
            Element_result.append(tmp_column)

        return pd.DataFrame(Element_result)
    
    def get_dayresult(self, step2_filename):
        summary, _, _ = self.step2(step2_filename)
               
        results = []
        for key, value in self.static_tables.items():
            result = [a / b for a, b in zip(value, summary)]
            results.append(result)

        print('\n', '#' * 10, 'DAY RESULTS', '#' * 10,'\n')
        dayresult = []
        for i, result in enumerate(results):
            i += 1
            print(f'Day {i} >> {mean(result)}')
            dayresult.append(mean(result))
            
        return (dayresult, summary)
    
    def batchproc(self, batch_filename):
    
        results = {}
        wb = openpyxl.load_workbook(batch_filename)
        print("Sheet Names", wb.sheetnames, '\n')
              
        ## CONTROL
        control = pd.read_excel(batch_filename, sheet_name='Control', header=0, engine="openpyxl")
        control_input = self.get_samples(control)

        control_average = []
        control_result = []
        number_cols = len(control_input.columns)
        number_rows = len(control_input)

        for i in range(number_cols):
            tmp_row = []
            for j in range(number_rows):
                tmp_row.append(control_input.iloc[j].iloc[i])
            control_average.append(mean(tmp_row))
            
        print("CONTROL AVERAGES\n", '-'*15+'\n', *["Day {}\t{}\n".format(di+1, control_average[di]) for di in range(len(control_average))], sep='')
        results = {'Control': pd.DataFrame(control_average)}
        
        
        ## OTHER SHEETS
        collist = list(set(wb.sheetnames)-set(['Control','Calc']))
        for x in collist:
            second_data = self.batch(batch_filename, x, control_average)
                        
            print("{} RESULT\n".format(x),  '-'*15+'\n', second_data.to_string(), '\n', sep='')
            results[x] = second_data.copy()

        print('Calculation is done.')
        
        return results
       
    
    def step2save(self, results_dict, outputdict, outputfile):
        
        print("Saving:")
        
        for filename, (summary, wells, tables) in results_dict.items():
            sheetname = '.'.join(filename.split('.')[:-1]).split('/')[-1]
            
            writer = pd.ExcelWriter(outputdict + 'step2_' + sheetname + '.xlsx', engine='xlsxwriter')
            print(outputdict + 'step2_' + sheetname + '.xlsx')
            wells.to_excel(writer, 'Summary')  
                        
            for sheet, table in tables.items():
                temp_t = table.rename(columns = { 
                    i: '' if i.startswith("Unnamed") else '.'.join([x for x in i.split('.') if not x.isdigit()]) 
                    for i in table.columns } )
                temp_t.to_excel(writer, sheet, index=False, startrow = 1)
            
            writer.save()
        
        if len(results_dict) > 1:
            print(outputdict + outputfile)
            writer = pd.ExcelWriter(outputdict + outputfile, engine='xlsxwriter')
            slist = []
            for filename, (summary, wells, _) in results_dict.items():
                sheetname = '.'.join(filename.split('.')[:-1]).split('/')[-1]
                wells.to_excel(writer, sheetname)  
                
                slist.append(summary.rename(sheetname))
            
            pd.concat(slist, axis=1).to_excel(writer, 'Averages')
            writer.save()
        
    def batchsave(self, results_dict, outputfile, outcontrol = False):
        writer = pd.ExcelWriter(outputfile, engine='xlsxwriter')
        print("Saving:")
        print(outputfile)
        
        for filename, results in results_dict.items():
            sheetn = filename.split('/')[-1]
            filex = sheetn.split('_')
            batchprocname = filex[1] if len(filex) >= 2 else 'b'
            
            collist = list(set(results.keys())-set(['Control','Calc']))
            control = results['Control']
            for x in collist:
                sheetname = batchprocname + '_' + x
                temp_res = results[x].copy()
                if outcontrol:
                    temp_res['Control'] = control
                temp_res.to_excel(writer, sheetname)
        
        writer.save()
    
    def mergeinputs1(self, filelist):
        outputfile = self.out_folder + 'step2_in.xlsx'
        writer = pd.ExcelWriter(outputfile, engine='xlsxwriter')
        mergefiles = {file: pd.read_excel(self.dir_step2 + file, skiprows=1, engine="openpyxl") for file in filelist}
            
        for file, table in mergefiles.items():
            sheetname = '.'.join(file.split('.')[:-1])
            table.to_excel(writer, sheetname, index=False)
            
        writer.save()
        
    def runs2(self, filename1):
        ## STEP 2
        #print('\n','-' * 36)
        #print('\n', '#' * 10, 'STEP 2 SELECT', '#' * 10,'\n')
        #for index, files in enumerate(self.files1):
        #    print(f'{index} -) {files}')
        #
        #print()
        #print('Choose an excel file: ', end='')
        #file_id = int(input())
        #filename1 = self.dir_step2 + self.files1[file_id]
        
        summary = self.step2(filename1)
        
        #step2 STORING
        outputfile = 's2.xlsx'
        self.step2save({filename1: summary}, self.out_folder, outputfile)
    
    def runbatch(self, filename2, outdir = None):
        ## BATCH
        #print('\n','-' * 36)
        #print('\n', '#' * 10, 'BATCH SELECT', '#' * 10,'\n')
        #for index, files in enumerate(self.files2):
        #    print(f'{index} -) {files}')
        #
        #print('\n', 'Choose an excel file: ', end='')
        #file_id = int(input())
        #filename2 = self.dir_batch + self.files2[file_id]
        
        results = self.batchproc(filename2)
        file_name = "Final_Result_" + os.path.basename(filename2)

        #batch STORING
        if outdir == None:
            outputfile = os.path.join(self.out_folder, file_name)
        else:
            outputfile = os.path.join(outdir, file_name)

        #rename if existed
        i = 1
        while True:
            if os.path.exists(outputfile):
                # add number extension to file name
                file_name = "Final_Result_" + os.path.basename(filename2).split('.')[0] + f'_{i}.xlsx'
                outputfile = self.out_folder + file_name
            else:
                break

        self.batchsave({filename2: results}, outputfile)
            
    def runalls2(self):
        ## STEP 2
        #print('\n','-' * 36)
        #print('\n', '#' * 10, 'STEP 2', '#' * 10,'\n')
        
        results1 = {}
        for index, files in enumerate(self.files1):
            print('-' * 36, '\n', f'{index+1} -) {files}\n', '-' * 36, '\n', sep='')
            results1[files] = self.step2(self.dir_step2 + files)
        
        ##step2 STORING
        outputfile = 's2_summary.xlsx'
        self.step2save(results1, self.out_folder, outputfile)
        #self.mergeinputs1(self.files1)
    
    def runallbatch(self):
        ## BATCH
        #print('\n','-' * 36)
        #print('\n', '#' * 10, 'BATCH', '#' * 10,'\n')
        
        results2 = {}
        for index, files in enumerate(self.files2):
            print('-' * 36, '\n', f'{index+1} -) {files}\n', '-' * 36, '\n', sep='')
            results = self.batchproc(self.dir_batch + files)
            results2[files] = results.copy()
        
        #batch STORING
        outputfile = self.out_folder + 'batch.xlsx'
        self.batchsave(results2, outputfile)

    ### VERSION 1.3 ADDED ###
    ### BY VINCENT L ###
    ### 2023 Mar 03 ###
    ### FOR MERGING SUMMARY DATA INTO BATCH INPUT ###

    def mergefiles(self, BATCH_DIR):

        # DEFINE SOME FUNCTIONS

        def append_df_to_excel(filename, df, sheet_name='Sheet1', startcol=None, startrow=None, col_sep = 0, row_sep = 0,
                       truncate_sheet=False, DISPLAY = False,
                       **to_excel_kwargs):
            # Excel file doesn't exist - saving and exiting
            if not os.path.isfile(filename):
                df.to_excel(
                    filename,
                    sheet_name=sheet_name, 
                    startcol=startcol if startcol is not None else 0, 
                    startrow=startrow if startrow is not None else 0,
                    **to_excel_kwargs)
                return
            
            if 'engine' in to_excel_kwargs:
                to_excel_kwargs.pop('engine')

            writer = pd.ExcelWriter(filename, engine='openpyxl', mode='a', if_sheet_exists='overlay')

            # try to open an existing workbook
            writer.workbook = openpyxl.load_workbook(filename)

            # get the last col in the existing Excel sheet
            # if it was not specified explicitly
            if startcol is None and sheet_name in writer.workbook.sheetnames:
                startcol = writer.workbook[sheet_name].max_column + col_sep

            if startrow is None and sheet_name in writer.workbook.sheetnames:
                startrow = writer.workbook[sheet_name].max_row + row_sep
            
            if startcol is None:
                startcol = 0

            if startrow is None:
                startrow = 0

            if DISPLAY:
                print('In file: ', os.path.basename(filename))
                print('Appended to column: ', startcol,', row: ', startrow)
            
            # write the dataframe to the existing sheet
            df.to_excel(writer, sheet_name, startcol=startcol, startrow=startrow, **to_excel_kwargs)

            # close workbook
            writer.close()

        def get_sheet_name(input_index):
            # if there is number inside input_index
            if re.search(r'\d', input_index):
                # remove the number
                suffix = re.sub(r'\d+$', '', input_index)
            else:
                suffix = 'Control'
            return suffix

        def remove_num(input_str):
            return re.sub(r'\d+$', '', input_str)

        # -------------------------------------------- 
        # MAIN 
        # --------------------------------------------
        # Received input is a directory that contains several data files (.xlsx) in separated folder day0, day1, day2, etc.
        # The task is to combine them in groups of Control, Substance A, Substance B, etc.
        # Calculate the Inhibition Average and Inhibition Average % for each group
        # Save these results in the same Merged excel file
        # --------------------------------------------
        
        # [TODO] HAVE SOME ADJUSTMENT TO DEAL WITH MISSING DATA

        print("MERGING FILES, PLEASE WAIT...")

        # Define OUTPUT_PATH
        output_path = os.path.join(BATCH_DIR, f'{os.path.basename(BATCH_DIR)}.xlsx')
        if os.path.exists(output_path):
            os.remove(output_path)

        # PATHS_DF = pd.DataFrame(columns=[f'day{i}' for i in range(0, self.days)])
        self.days_list = os.listdir(BATCH_DIR)
        PATHS_DF = pd.DataFrame(columns=self.days_list)

        # find all files in the batch directory, any level deep, with .xlsx extension
        data_points = []
        for file_path in Path(BATCH_DIR).rglob('*.xlsx'):
            # get file base name without extension
            file_basename = file_path.stem
            file_basename = str(file_basename).upper()
            if not re.search(r'\d', file_basename):
                file_basename = 'Control'
            # find in the file_path, see which day it belongs to
            for day in self.days_list:
                if day in str(file_path):
                    # add the file_path to the dataframe
                    PATHS_DF.loc[file_basename, day] = file_path
                    data_points.append(file_basename)
                    break 
        # print(PATHS_DF)

        AVG_DF_DICT = {}
        startcol = 0
        for day in PATHS_DF.columns:
            temp_dict = {}
            lastrow = 0
            # interate through index
            for category in PATHS_DF.index:
                file_path = PATHS_DF.loc[category, day]

                # if the file_path is Nan, raise Exception
                if pd.isna(file_path):
                    raise Exception(f'Missing file for {category} on {day}')
                
                try:
                    df = pd.read_excel(file_path, sheet_name='Summary', header=0, engine="openpyxl")
                except ValueError:
                    print(f'{file_path} has no Worksheet with the name Summary')
                    
                # if df is empty (no data), raise Exception
                if df.empty:
                    raise Exception(f'Empty file for {category} on {day}')
                
                # if index has number 3 (last set of wells), remove the last 3 rows of the df
                if re.search(r'3$', category):
                    df = df[:-3]

                # if index has no number (Control) or has number 1 (First set of wells) in it, set following inputs
                if not re.search(r'\d', category) or re.search(r'1$', category):
                    startrow = 0
                    header = True
                    temp_dict[remove_num(category)] = df['Average'].tolist()
                else:
                    if re.search(r'2$', category):
                        startrow = lastrow + 1
                    else:
                        startrow = lastrow
                    header = False
                    temp_dict[remove_num(category)].extend(df['Average'].tolist())

                # update lastrow
                lastrow = startrow + len(df)

                # Append the df to the excel file
                sheet_name = get_sheet_name(category)
                append_df_to_excel(output_path, df, sheet_name=sheet_name, startrow = startrow, startcol = startcol, index=False, header = header)

            if day == 'Day_0':
                for key in temp_dict.keys():
                    AVG_DF_DICT[key] = pd.DataFrame(columns=self.days_list)
            for key in temp_dict.keys():
                AVG_DF_DICT[key][day] = temp_dict[key]

            df_width = len(df.columns)
            startcol += df_width

        print('Calculating and adding Inhibition Average to excel files...')
        # append the average df to the excel file
        for key in AVG_DF_DICT.keys():
            append_df_to_excel(output_path, AVG_DF_DICT[key], sheet_name=key, startrow = None, startcol = 1, row_sep = 1, index=False, header = False)


        print('Calculating and adding Inhibition Average % to excel files...')
        # calculate the compare df
        COMPARE_DF_DICT = {}
        for key in AVG_DF_DICT.keys():
            # compare_df = avg_df.iloc[:,1:].div(avg_df.iloc[:,0], axis=0)
            COMPARE_DF_DICT[key] = AVG_DF_DICT[key].iloc[:,1:].div(AVG_DF_DICT[key].iloc[:,0], axis=0)
        
        # append the compare df to the excel file
        for key in COMPARE_DF_DICT.keys():
            append_df_to_excel(output_path, COMPARE_DF_DICT[key], sheet_name=key, startrow = None, startcol = 2, row_sep = 1, index=False, header = False)

        print()
        print('Merged files were created!')
        print()
        time.sleep(2)

    def combine(self, BATCH_DIR):
        # --------------------------------------------
        pass

        


if __name__ == '__main__':
    calc = Calculate()
    calc.run()
