import re
import os
import openpyxl
import pandas as pd



def append_df_to_excel(filename, df, sheet_name='Sheet1', startcol=None, startrow=None, col_sep = 0, row_sep = 0,
                       truncate_sheet=False, DISPLAY = False,
                       **to_excel_kwargs):
    # Excel file doesn't exist - saving and exiting
    if not os.path.isfile(filename):
        df.to_excel(
            filename,
            sheet_name=sheet_name, 
            startcol=startcol if startcol is not None else 0, 
            startrow=startrow if startrow is not None else 0,
            **to_excel_kwargs)
        return
    
    if 'engine' in to_excel_kwargs:
        to_excel_kwargs.pop('engine')

    writer = pd.ExcelWriter(filename, engine='openpyxl', mode='a', if_sheet_exists='overlay')

    # try to open an existing workbook
    writer.workbook = openpyxl.load_workbook(filename)

    # get the last col in the existing Excel sheet
    # if it was not specified explicitly
    if startcol is None and sheet_name in writer.workbook.sheetnames:
        startcol = writer.workbook[sheet_name].max_column + col_sep

    if startrow is None and sheet_name in writer.workbook.sheetnames:
        startrow = writer.workbook[sheet_name].max_row + row_sep
    
    if startcol is None:
        startcol = 0

    if startrow is None:
        startrow = 0

    if DISPLAY:
        print('In file: ', os.path.basename(filename))
        print('Appended to column: ', startcol,', row: ', startrow)
    
    # write the dataframe to the existing sheet
    df.to_excel(writer, sheet_name, startcol=startcol, startrow=startrow, **to_excel_kwargs)

    # close workbook
    writer.close()

def get_sheet_name(input_index):
    # if there is number inside input_index
    if re.search(r'\d', input_index):
        # remove the number
        prefix = re.sub(r'\d+$', '', input_index)
    else:
        prefix = 'Control'
    return prefix

def remove_num(input_str):
        return re.sub(r'\d+$', '', input_str)

def splitter(input_str):
    if re.search(r'\d', input_str):
        string_part = re.sub(r'\d+$', '', input_str)
        num_part = re.search(r'\d+$', input_str).group()
        num_part = int(num_part)
    else:
        string_part = input_str
        num_part = None
    return string_part, num_part

def input_with_type(prompt, type_, default_value = None):
    if default_value is not None:
        prompt = prompt + ' (default: ' + str(default_value) + '), press Enter to use default : '
        user_input = input(prompt)
        if user_input == '':
            return default_value
        else:
            while True:
                try:
                    return type_(input(prompt))
                except ValueError:
                    print(f"Please enter a {type_} value.")
    else:
        prompt += ' : '
        while True:
            try:
                return type_(input(prompt))
            except ValueError:
                print(f"Please enter a {type_} value.")